import json
from datetime import datetime, timedelta

import pystac
import pytest
import requests
from dateutil.tz import gettz, tzutc
from pystac_client import Client
from pystac_client.item_search import ItemSearch

from .helpers import STAC_URLS, read_data_file

SEARCH_URL = f"{STAC_URLS['PLANETARY-COMPUTER']}/search"
INTERSECTS_EXAMPLE = {
    'type':
    'Polygon',
    'coordinates': [[[-73.21, 43.99], [-73.21, 44.05], [-73.12, 44.05], [-73.12, 43.99],
                     [-73.21, 43.99]]]
}

ITEM_EXAMPLE = {"collections": "io-lulc", "ids": "60U-2020"}


@pytest.mark.skip(reason="Performance testing skipped in normal test run")
class TestItemPerformance:
    @pytest.fixture(scope='function')
    def single_href(self):
        item_href = "https://planetarycomputer.microsoft.com/api/stac/v1/collections/{collections}/items/{ids}".format(
            collections=ITEM_EXAMPLE['collections'], ids=ITEM_EXAMPLE['ids'])
        return item_href

    def test_requests(self, benchmark, single_href):
        response = benchmark(requests.get, single_href)
        assert response.status_code == 200

        assert response.json()['id'] == ITEM_EXAMPLE["ids"]

    def test_single_item(self, benchmark, single_href):
        item = benchmark(pystac.Item.from_file, single_href)

        assert item.id == ITEM_EXAMPLE["ids"]

    def test_single_item_search(self, benchmark, single_href):
        search = ItemSearch(url=SEARCH_URL, **ITEM_EXAMPLE)

        item_collection = benchmark(search.get_all_items)

        assert len(item_collection.items) == 1
        assert item_collection.items[0].id == ITEM_EXAMPLE["ids"]


class TestItemSearchParams:
    @pytest.fixture(scope='function')
    def sample_client(self):
        api_content = read_data_file('planetary-computer-root.json', parse_json=True)
        return Client.from_dict(api_content)

    def test_tuple_bbox(self):
        # Tuple input
        search = ItemSearch(url=SEARCH_URL, bbox=(-104.5, 44.0, -104.0, 45.0))
        assert search._parameters['bbox'] == (-104.5, 44.0, -104.0, 45.0)

    def test_list_bbox(self):
        # List input
        search = ItemSearch(url=SEARCH_URL, bbox=[-104.5, 44.0, -104.0, 45.0])
        assert search._parameters['bbox'] == (-104.5, 44.0, -104.0, 45.0)

    def test_string_bbox(self):
        # String Input
        search = ItemSearch(url=SEARCH_URL, bbox='-104.5,44.0,-104.0,45.0')
        assert search._parameters['bbox'] == (-104.5, 44.0, -104.0, 45.0)

    def test_generator_bbox(self):
        # Generator Input
        def bboxer():
            yield from [-104.5, 44.0, -104.0, 45.0]

        search = ItemSearch(url=SEARCH_URL, bbox=bboxer())
        assert search._parameters['bbox'] == (-104.5, 44.0, -104.0, 45.0)

    def test_single_string_datetime(self):
        # Single timestamp input
        search = ItemSearch(url=SEARCH_URL, datetime='2020-02-01T00:00:00Z')
        assert search._parameters['datetime'] == '2020-02-01T00:00:00Z'

    def test_range_string_datetime(self):
        # Timestamp range input
        search = ItemSearch(url=SEARCH_URL, datetime='2020-02-01T00:00:00Z/2020-02-02T00:00:00Z')
        assert search._parameters['datetime'] == '2020-02-01T00:00:00Z/2020-02-02T00:00:00Z'

    def test_list_of_strings_datetime(self):
        # Timestamp list input
        search = ItemSearch(url=SEARCH_URL,
                            datetime=['2020-02-01T00:00:00Z', '2020-02-02T00:00:00Z'])
        assert search._parameters['datetime'] == '2020-02-01T00:00:00Z/2020-02-02T00:00:00Z'

    def test_open_range_string_datetime(self):
        # Open timestamp range input
        search = ItemSearch(url=SEARCH_URL, datetime='2020-02-01T00:00:00Z/..')
        assert search._parameters['datetime'] == '2020-02-01T00:00:00Z/..'

    def test_single_datetime_object(self):
        start = datetime(2020, 2, 1, 0, 0, 0, tzinfo=tzutc())

        # Single datetime input
        search = ItemSearch(url=SEARCH_URL, datetime=start)
        assert search._parameters['datetime'] == '2020-02-01T00:00:00Z'

    def test_list_of_datetimes(self):
        start = datetime(2020, 2, 1, 0, 0, 0, tzinfo=tzutc())
        end = datetime(2020, 2, 2, 0, 0, 0, tzinfo=tzutc())

        # Datetime range input
        search = ItemSearch(url=SEARCH_URL, datetime=[start, end])
        assert search._parameters['datetime'] == '2020-02-01T00:00:00Z/2020-02-02T00:00:00Z'

    def test_open_list_of_datetimes(self):
        start = datetime(2020, 2, 1, 0, 0, 0, tzinfo=tzutc())

        # Open datetime range input
        search = ItemSearch(url=SEARCH_URL, datetime=(start, None))
        assert search._parameters['datetime'] == '2020-02-01T00:00:00Z/..'

    def test_localized_datetime_converted_to_utc(self):
        # Localized datetime input (should be converted to UTC)
        start_localized = datetime(2020, 2, 1, 0, 0, 0, tzinfo=gettz('US/Eastern'))
        search = ItemSearch(url=SEARCH_URL, datetime=start_localized)
        assert search._parameters['datetime'] == '2020-02-01T05:00:00Z'

    def test_single_year(self):
        search = ItemSearch(url=SEARCH_URL, datetime='2020')
        assert search._parameters['datetime'] == "2020-01-01T00:00:00Z/2020-12-31T23:59:59Z"

    def test_range_of_years(self):
        search = ItemSearch(url=SEARCH_URL, datetime='2019/2020')
        assert search._parameters['datetime'] == "2019-01-01T00:00:00Z/2020-12-31T23:59:59Z"

    def test_single_month(self):
        search = ItemSearch(url=SEARCH_URL, datetime='2020-06')
        assert search._parameters['datetime'] == "2020-06-01T00:00:00Z/2020-06-30T23:59:59Z"

    def test_range_of_months(self):
        search = ItemSearch(url=SEARCH_URL, datetime='2020-04/2020-06')
        assert search._parameters['datetime'] == "2020-04-01T00:00:00Z/2020-06-30T23:59:59Z"

    def test_single_date(self):
        search = ItemSearch(url=SEARCH_URL, datetime='2020-06-10')
        assert search._parameters['datetime'] == "2020-06-10T00:00:00Z/2020-06-10T23:59:59Z"

    def test_range_of_dates(self):
        search = ItemSearch(url=SEARCH_URL, datetime='2020-06-10/2020-06-20')
        assert search._parameters['datetime'] == "2020-06-10T00:00:00Z/2020-06-20T23:59:59Z"

    def test_mixed_simple_date_strings(self):
        search = ItemSearch(url=SEARCH_URL, datetime="2019/2020-06-10")
        assert search._parameters['datetime'] == "2019-01-01T00:00:00Z/2020-06-10T23:59:59Z"

    def test_time(self):
        search = ItemSearch(url=SEARCH_URL, datetime="2019-01-01T00:00:00Z/2019-01-01T00:12:00")
        assert search._parameters['datetime'] == "2019-01-01T00:00:00Z/2019-01-01T00:12:00Z"

    def test_many_datetimes(self):
        datetimes = [
            "1985-04-12T23:20:50.52Z",
            "1996-12-19T16:39:57-08:00",
            "1990-12-31T23:59:60Z",
            "1990-12-31T15:59:60-08:00",
            "1937-01-01T12:00:27.87+01:00",
            "1985-04-12T23:20:50.52Z",
            "1937-01-01T12:00:27.8710+01:00",
            "1937-01-01T12:00:27.8+01:00",
            "1937-01-01T12:00:27.8Z",
            "1985-04-12t23:20:50.5202020z",
            "2020-07-23T00:00:00Z",
            "2020-07-23T00:00:00.0Z",
            "2020-07-23T00:00:00.01Z",
            "2020-07-23T00:00:00.012Z",
            "2020-07-23T00:00:00.0123Z",
            "2020-07-23T00:00:00.01234Z",
            "2020-07-23T00:00:00.012345Z",
            "2020-07-23T00:00:00.000Z",
            "2020-07-23T00:00:00.000+03:00",
            "2020-07-23T00:00:00+03:00",
            "2020-07-23T00:00:00.000+03:00",
            "2020-07-23T00:00:00.000z",
        ]
        for date_time in datetimes:
            ItemSearch(url=SEARCH_URL, datetime=date_time)

    def test_three_datetimes(self):
        start = datetime(2020, 2, 1, 0, 0, 0, tzinfo=tzutc())
        middle = datetime(2020, 2, 2, 0, 0, 0, tzinfo=tzutc())
        end = datetime(2020, 2, 3, 0, 0, 0, tzinfo=tzutc())

        with pytest.raises(Exception):
            ItemSearch(url=SEARCH_URL, datetime=[start, middle, end])

    def test_single_collection_string(self):
        # Single ID string
        search = ItemSearch(url=SEARCH_URL, collections='naip')
        assert search._parameters['collections'] == ('naip', )

    def test_multiple_collection_string(self):
        # Comma-separated ID string
        search = ItemSearch(url=SEARCH_URL, collections='naip,landsat8_l1tp')
        assert search._parameters['collections'] == ('naip', 'landsat8_l1tp')

    def test_list_of_collection_strings(self):
        # List of ID strings
        search = ItemSearch(url=SEARCH_URL, collections=['naip', 'landsat8_l1tp'])
        assert search._parameters['collections'] == ('naip', 'landsat8_l1tp')

    def test_generator_of_collection_strings(self):
        # Generator of ID strings
        def collectioner():
            yield from ['naip', 'landsat8_l1tp']

        search = ItemSearch(url=SEARCH_URL, collections=collectioner())
        assert search._parameters['collections'] == ('naip', 'landsat8_l1tp')

    def test_single_id_string(self):
        # Single ID
        search = ItemSearch(url=SEARCH_URL, ids='m_3510836_se_12_060_20180508_20190331')
        assert search._parameters['ids'] == ('m_3510836_se_12_060_20180508_20190331', )

    def test_multiple_id_string(self):
        # Comma-separated ID string
        search = ItemSearch(
            url=SEARCH_URL,
            ids='m_3510836_se_12_060_20180508_20190331,m_3510840_se_12_060_20180504_20190331')
        assert search._parameters['ids'] == ('m_3510836_se_12_060_20180508_20190331',
                                             'm_3510840_se_12_060_20180504_20190331')

    def test_list_of_id_strings(self):
        # List of IDs
        search = ItemSearch(
            url=SEARCH_URL,
            ids=['m_3510836_se_12_060_20180508_20190331', 'm_3510840_se_12_060_20180504_20190331'])
        assert search._parameters['ids'] == ('m_3510836_se_12_060_20180508_20190331',
                                             'm_3510840_se_12_060_20180504_20190331')

    def test_generator_of_id_string(self):
        # Generator of IDs
        def ids():
            yield from [
                'm_3510836_se_12_060_20180508_20190331', 'm_3510840_se_12_060_20180504_20190331'
            ]

        search = ItemSearch(url=SEARCH_URL, ids=ids())
        assert search._parameters['ids'] == ('m_3510836_se_12_060_20180508_20190331',
                                             'm_3510840_se_12_060_20180504_20190331')

    def test_intersects_dict(self):
        # Dict input
        search = ItemSearch(url=SEARCH_URL, intersects=INTERSECTS_EXAMPLE)
        assert search._parameters['intersects'] == INTERSECTS_EXAMPLE

    def test_intersects_json_string(self):
        # JSON string input
        search = ItemSearch(url=SEARCH_URL, intersects=json.dumps(INTERSECTS_EXAMPLE))
        assert search._parameters['intersects'] == INTERSECTS_EXAMPLE


class TestItemSearch:
    @pytest.fixture(scope='function')
    def astraea_api(self):
        api_content = read_data_file('astraea_api.json', parse_json=True)
        return Client.from_dict(api_content)

    def test_method(self):
        # Default method should be POST...
        search = ItemSearch(url=SEARCH_URL)
        assert search.method == 'POST'

        # "method" argument should take precedence over presence of "intersects"
        search = ItemSearch(url=SEARCH_URL, method='GET', intersects=INTERSECTS_EXAMPLE)
        assert search.method == 'GET'

    @pytest.mark.vcr
    def test_results(self):
        search = ItemSearch(
            url=SEARCH_URL,
            collections='naip',
            max_items=20,
            limit=10,
        )
        results = search.get_items()

        assert all(isinstance(item, pystac.Item) for item in results)

    @pytest.mark.vcr
    def test_ids_results(self):
        ids = [
            'S2B_MSIL2A_20210610T115639_N0212_R066_T33XXG_20210613T185024.SAFE',
            'fl_m_2608004_nw_17_060_20191215_20200113'
        ]
        search = ItemSearch(
            url=SEARCH_URL,
            ids=ids,
        )
        results = list(search.get_items())

        assert len(results) == 1
        assert all(item.id in ids for item in results)

    @pytest.mark.vcr
    def test_datetime_results(self):
        # Datetime range string
        datetime_ = '2019-01-01T00:00:01Z/2019-01-01T00:00:10Z'
        search = ItemSearch(url=SEARCH_URL, datetime=datetime_)
        results = list(search.get_items())
        assert len(results) == 33

        min_datetime = datetime(2019, 1, 1, 0, 0, 1, tzinfo=tzutc())
        max_datetime = datetime(2019, 1, 1, 0, 0, 10, tzinfo=tzutc())
        search = ItemSearch(url=SEARCH_URL, datetime=(min_datetime, max_datetime))
        results = search.get_items()
        assert all(min_datetime <= item.datetime <= (max_datetime + timedelta(seconds=1))
                   for item in results)

    @pytest.mark.vcr
    def test_intersects_results(self):
        # GeoJSON-like dict
        intersects_dict = {
            'type':
            'Polygon',
            'coordinates': [[[-73.21, 43.99], [-73.21, 44.05], [-73.12, 44.05], [-73.12, 43.99],
                             [-73.21, 43.99]]]
        }
        search = ItemSearch(url=SEARCH_URL, intersects=intersects_dict, collections='naip')
        results = list(search.get_items())
        assert len(results) == 30

        # Geo-interface object
        class MockGeoObject:
            __geo_interface__ = intersects_dict

        intersects_obj = MockGeoObject()
        search = ItemSearch(url=SEARCH_URL, intersects=intersects_obj, collections='naip')
        results = search.get_items()
        assert all(isinstance(item, pystac.Item) for item in results)

    @pytest.mark.vcr
    def test_result_paging(self):
        search = ItemSearch(
            url=SEARCH_URL,
            bbox=(-73.21, 43.99, -73.12, 44.05),
            collections='naip',
            limit=10,
            max_items=20,
        )

        # Check that the current page changes on the ItemSearch instance when a new page is requested
        pages = list(search.get_item_collections())

        assert pages[1] != pages[2]
        assert pages[1].items != pages[2].items

    @pytest.mark.vcr
    def test_get_all_items(self):
        search = ItemSearch(
            url=SEARCH_URL,
            bbox=(-73.21, 43.99, -73.12, 44.05),
            collections='naip',
            limit=10,
            max_items=20,
        )
        item_collection = search.get_all_items()
        assert len(item_collection.items) == 20


class TestItemSearchQuery:
    def test_query_shortcut_syntax(self):
        search = ItemSearch(url=SEARCH_URL,
                            bbox=(-73.21, 43.99, -73.12, 44.05),
                            query=["gsd=10"],
                            max_items=1)
        items1 = list(search.get_items())

        search = ItemSearch(url=SEARCH_URL,
                            bbox=(-73.21, 43.99, -73.12, 44.05),
                            query={"gsd": {
                                "eq": 10
                            }},
                            max_items=1)
        items2 = list(search.get_items())

        assert (len(items1) == 1)
        assert (len(items2) == 1)
        assert (items1[0].id == items2[0].id)



from typing import (Iterable, TYPE_CHECKING)

import pystac
from pystac_client.item_search import ItemSearch

if TYPE_CHECKING:
    from pystac.item import Item as Item_Type


class CollectionClient(pystac.Collection):
    def __repr__(self):
        return '<CollectionClient id={}>'.format(self.id)

    def get_items(self) -> Iterable["Item_Type"]:
        """Return all items in this Collection.

        If the Collection contains a link of with a `rel` value of `items`, that link will be
        used to iterate through items. Otherwise, the default PySTAC behavior is assumed.

        Return:
            Iterable[Item]: Generator of items whose parent is this catalog.
        """

        link = self.get_single_link('items')
        if link is not None:
            search = ItemSearch(link.href, method='GET')
            yield from search.get_items()
        else:
            yield from super().get_items()



# flake8: noqa

from pystac_client.version import __version__
from pystac_client.item_search import ItemSearch
from pystac_client.client import Client
from pystac_client.collection_client import CollectionClient
from pystac_client.conformance import ConformanceClasses



from copy import deepcopy
import json
import logging
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    TYPE_CHECKING,
    Union,
)
from urllib.parse import urlparse
import re
from requests import Request, Session

import pystac
from pystac.link import Link
from pystac.serialization import (
    merge_common_properties,
    identify_stac_object_type,
    identify_stac_object,
    migrate_to_latest,
)
from pystac.stac_io import DefaultStacIO

import pystac_client
from .exceptions import APIError
from .conformance import ConformanceClasses, CONFORMANCE_URIS

if TYPE_CHECKING:
    from pystac.stac_object import STACObject as STACObject_Type
    from pystac.catalog import Catalog as Catalog_Type

logger = logging.getLogger(__name__)


class StacApiIO(DefaultStacIO):
    def __init__(self, headers: Optional[Dict] = None, conformance: Optional[List[str]] = None):
        """Initialize class for API IO

        Args:
            headers : Optional dictionary of headers to include in all requests

        Return:
            StacApiIO : StacApiIO instance
        """
        # TODO - this should super() to parent class
        self.session = Session()
        self.session.headers.update(headers or {})

        self._conformance = conformance

    def read_text(self,
                  source: Union[str, Link],
                  *args: Any,
                  parameters: Optional[dict] = {},
                  **kwargs: Any) -> str:
        """Read text from the given URI.

        Overwrites the default method for reading text from a URL or file to allow :class:`urllib.request.Request`
        instances as input. This method also raises any :exc:`urllib.error.HTTPError` exceptions rather than catching
        them to allow us to handle different response status codes as needed.
        """
        if isinstance(source, str):
            href = source
            if bool(urlparse(href).scheme):
                return self.request(href, *args, parameters=parameters, **kwargs)
            else:
                with open(href) as f:
                    href_contents = f.read()
                return href_contents
        elif isinstance(source, Link):
            link = source.to_dict()
            href = link['href']
            # get headers and body from Link and add to request from simple stac resolver
            merge = bool(link.get('merge', False))

            # If the link object includes a "method" property, use that. If not fall back to 'GET'.
            method = link.get('method', 'GET')
            # If the link object includes a "headers" property, use that and respect the "merge" property.
            headers = link.get('headers', None)

            # If "POST" use the body object that and respect the "merge" property.
            link_body = link.get('body', {})
            if method == 'POST':
                parameters = {**parameters, **link_body} if merge else link_body
            else:
                # parameters are already in the link href
                parameters = {}
            return self.request(href, *args, method=method, headers=headers, parameters=parameters)

    def request(self,
                href: str,
                method: Optional[str] = 'GET',
                headers: Optional[dict] = {},
                parameters: Optional[dict] = {}) -> str:
        """Makes a request to an http endpoint

        Args:
            href (str): The request URL
            method (Optional[str], optional): The http method to use, 'GET' or 'POST'. Defaults to 'GET'.
            headers (Optional[dict], optional): Additional headers to include in request. Defaults to {}.
            parameters (Optional[dict], optional): parameters to send with request. Defaults to {}.

        Raises:
            APIError: raised if the server returns an error response

        Return:
            str: The decoded response from the endpoint
        """
        if method == 'POST':
            request = Request(method=method, url=href, headers=headers, json=parameters)
        else:
            params = deepcopy(parameters)
            if 'intersects' in params:
                params['intersects'] = json.dumps(params['intersects'])
            request = Request(method=method, url=href, headers=headers, params=params)
        try:
            prepped = self.session.prepare_request(request)
            msg = f"{prepped.method} {prepped.url} Headers: {prepped.headers}"
            if method == 'POST':
                msg += f" Payload: {json.dumps(request.json)}"
            logger.debug(msg)
            resp = self.session.send(prepped)
            if resp.status_code != 200:
                raise APIError(resp.text)
            return resp.content.decode("utf-8")
        except Exception as err:
            raise APIError(str(err))

    def write_text_to_href(self, href: str, *args: Any, **kwargs: Any) -> None:
        if bool(urlparse(href).scheme):
            raise APIError("Transactions not supported")
        else:
            return super().write_text_to_href(*args, **kwargs)

    def stac_object_from_dict(
        self,
        d: Dict[str, Any],
        href: Optional[str] = None,
        root: Optional["Catalog_Type"] = None,
        preserve_dict: bool = True,
    ) -> "STACObject_Type":
        """Deserializes a :class:`~pystac.STACObject` sub-class instance from a dictionary.

        Args:
            d : The dictionary to deserialize
            href : Optional href to associate with the STAC object
            root : Optional root :class:`~pystac.Catalog` to associate with the
                STAC object.
            preserve_dict: If ``False``, the dict parameter ``d`` may be modified
                during this method call. Otherwise the dict is not mutated.
                Defaults to ``True``, which results results in a deepcopy of the
                parameter. Set to ``False`` when possible to avoid the performance
                hit of a deepcopy.
        """
        if identify_stac_object_type(d) == pystac.STACObjectType.ITEM:
            collection_cache = None
            if root is not None:
                collection_cache = root._resolved_objects.as_collection_cache()

            # Merge common properties in case this is an older STAC object.
            merge_common_properties(d, json_href=href, collection_cache=collection_cache)

        info = identify_stac_object(d)
        d = migrate_to_latest(d, info)

        if info.object_type == pystac.STACObjectType.CATALOG:
            result = pystac_client.Client.from_dict(d,
                                                    href=href,
                                                    root=root,
                                                    migrate=False,
                                                    preserve_dict=preserve_dict)
            result._stac_io = self
            return result

        if info.object_type == pystac.STACObjectType.COLLECTION:
            return pystac_client.CollectionClient.from_dict(d,
                                                            href=href,
                                                            root=root,
                                                            migrate=False,
                                                            preserve_dict=preserve_dict)

        if info.object_type == pystac.STACObjectType.ITEM:
            return pystac.Item.from_dict(d,
                                         href=href,
                                         root=root,
                                         migrate=False,
                                         preserve_dict=preserve_dict)

        raise ValueError(f"Unknown STAC object type {info.object_type}")

    def get_pages(self, url, method='GET', parameters={}) -> Iterator[Dict]:
        """Iterator that yields dictionaries for each page at a STAC paging endpoint, e.g., /collections, /search

        Return:
            Dict : JSON content from a single page
        """
        page = self.read_json(url, method=method, parameters=parameters)
        yield page

        next_link = next((link for link in page.get('links', []) if link['rel'] == 'next'), None)
        while next_link:
            link = Link.from_dict(next_link)
            page = self.read_json(link, parameters=parameters)
            yield page

            # get the next link and make the next request
            next_link = next((link for link in page.get('links', []) if link['rel'] == 'next'),
                             None)

    def assert_conforms_to(self, conformance_class: ConformanceClasses):
        """Whether the API conforms to the given standard. This method only checks against the ``"conformsTo"``
        property from the API landing page and does not make any additional calls to a ``/conformance`` endpoint
        even if the API provides such an endpoint.

        Args:
            key: The ``ConformanceClasses`` key to check conformance against.

        Return:
            bool: Indicates if the API conforms to the given spec or URI.
        """
        if not self.conforms_to(conformance_class):
            raise NotImplementedError(f"{conformance_class} not supported")
        else:
            return True

    def conforms_to(self, conformance_class: ConformanceClasses) -> bool:
        """Whether the API conforms to the given standard. This method only checks against the ``"conformsTo"``
        property from the API landing page and does not make any additional calls to a ``/conformance`` endpoint
        even if the API provides such an endpoint.

        Args:
            key : The ``ConformanceClasses`` key to check conformance against.

        Return:
            bool: Indicates if the API conforms to the given spec or URI.
        """

        # Conformance of None means ignore all conformance as opposed to an
        #  empty array which would indicate the API conforms to nothing
        if self._conformance is None:
            return True

        class_regex = CONFORMANCE_URIS.get(conformance_class.name, None)

        if class_regex is None:
            raise Exception(f"Invalid conformance class {conformance_class}")

        pattern = re.compile(class_regex)

        if not any(re.match(pattern, uri) for uri in self._conformance):
            return False

        return True

    def set_conformance(self, conformance: Optional[List[str]]) -> None:
        """Sets (or clears) the conformances for this StacIO."""
        self._conformance = conformance



import pytest

from pystac_client.exceptions import APIError
from pystac_client.stac_api_io import StacApiIO
from .helpers import STAC_URLS


class TestSTAC_IOOverride:
    @pytest.mark.vcr
    def test_request_input(self):
        stac_api_io = StacApiIO()
        response = stac_api_io.read_text(STAC_URLS['PLANETARY-COMPUTER'])
        assert isinstance(response, str)

    @pytest.mark.vcr
    def test_str_input(self):
        stac_api_io = StacApiIO()
        response = stac_api_io.read_text(STAC_URLS['PLANETARY-COMPUTER'])

        assert isinstance(response, str)

    @pytest.mark.vcr
    def test_http_error(self):
        stac_api_io = StacApiIO()
        # Attempt to access an authenticated endpoint
        with pytest.raises(APIError) as excinfo:
            stac_api_io.read_text(f"{STAC_URLS['MLHUB']}/search")

        assert isinstance(excinfo.value, APIError)

    def test_local_file(self, tmp_path):
        stac_api_io = StacApiIO()
        test_file = tmp_path / 'test.txt'
        with test_file.open('w') as dst:
            dst.write('Hi there!')

        response = stac_api_io.read_text(str(test_file))

        assert response == 'Hi there!'



import pytest

from pystac_client import CollectionClient
from pystac_client.client import Client

from .helpers import STAC_URLS


class TestCollectionClient:
    @pytest.mark.vcr
    def test_instance(self):
        client = Client.open(STAC_URLS['PLANETARY-COMPUTER'])
        collection = client.get_collection('aster-l1t')

        assert isinstance(collection, CollectionClient)
        assert str(collection) == '<CollectionClient id=aster-l1t>'

    @pytest.mark.vcr
    def test_get_items(self):
        client = Client.open(STAC_URLS['PLANETARY-COMPUTER'])
        collection = client.get_collection('aster-l1t')
        for item in collection.get_items():
            assert (item.collection_id == collection.id)
            return



from enum import Enum
import re


class ConformanceClasses(Enum):
    """Enumeration class for Conformance Classes

    """

    stac_prefix = re.escape("https://api.stacspec.org/v1.0.")

    # defined conformance classes regexes
    CORE = fr"{stac_prefix}(.*){re.escape('/core')}"
    ITEM_SEARCH = fr"{stac_prefix}(.*){re.escape('/item-search')}"
    CONTEXT = fr"{stac_prefix}(.*){re.escape('/item-search#context')}"
    FIELDS = fr"{stac_prefix}(.*){re.escape('/item-search#fields')}"
    SORT = fr"{stac_prefix}(.*){re.escape('/item-search#sort')}"
    QUERY = fr"{stac_prefix}(.*){re.escape('/item-search#query')}"
    FILTER = fr"{stac_prefix}(.*){re.escape('/item-search#filter')}"
    COLLECTIONS = re.escape("http://www.opengis.net/spec/ogcapi-features-1/1.0/conf/oas30")


CONFORMANCE_URIS = {c.name: c.value for c in ConformanceClasses}



from datetime import datetime

from dateutil.tz import tzutc
import pystac
import pytest

from pystac_client import Client
from pystac_client.conformance import ConformanceClasses

from .helpers import STAC_URLS, TEST_DATA


class TestAPI:
    @pytest.mark.vcr
    def test_instance(self):
        api = Client.open(STAC_URLS['PLANETARY-COMPUTER'])

        # An API instance is also a Catalog instance
        assert isinstance(api, pystac.Catalog)

        assert str(api) == '<Client id=microsoft-pc>'

    @pytest.mark.vcr
    def test_links(self):
        api = Client.open(STAC_URLS['PLANETARY-COMPUTER'])

        # Should be able to get collections via links as with a typical PySTAC Catalog
        collection_links = api.get_links('child')
        assert len(collection_links) > 0

        collections = list(api.get_collections())
        assert len(collection_links) == len(collections)

        first_collection = api.get_single_link('child').resolve_stac_object(root=api).target
        assert isinstance(first_collection, pystac.Collection)

    def test_spec_conformance(self):
        """Testing conformance against a ConformanceClass should allow APIs using legacy URIs to pass."""
        client = Client.from_file(str(TEST_DATA / 'planetary-computer-root.json'))

        # Set conformsTo URIs to conform with STAC API - Core using official URI
        client._stac_io._conformance = ['https://api.stacspec.org/v1.0.0-beta.1/core']

        assert client._stac_io.assert_conforms_to(ConformanceClasses.CORE)

    @pytest.mark.vcr
    def test_no_conformance(self):
        """Should raise a NotImplementedError if no conformance info can be found. Luckily, the test API doesn't publish
        a "conformance" link so we can just remove the "conformsTo" attribute to test this."""
        client = Client.from_file(str(TEST_DATA / 'planetary-computer-root.json'))
        client._stac_io._conformance = []

        with pytest.raises(NotImplementedError):
            client._stac_io.assert_conforms_to(ConformanceClasses.CORE)

        with pytest.raises(NotImplementedError):
            client._stac_io.assert_conforms_to(ConformanceClasses.ITEM_SEARCH)

    @pytest.mark.vcr
    def test_no_stac_core_conformance(self):
        """Should raise a NotImplementedError if the API does not conform to the STAC API - Core spec."""
        client = Client.from_file(str(TEST_DATA / 'planetary-computer-root.json'))
        client._stac_io._conformance = client._stac_io._conformance[1:]

        with pytest.raises(NotImplementedError):
            client._stac_io.assert_conforms_to(ConformanceClasses.CORE)

        assert client._stac_io.assert_conforms_to(ConformanceClasses.ITEM_SEARCH)

    @pytest.mark.vcr
    def test_from_file(self):
        api = Client.from_file(STAC_URLS['PLANETARY-COMPUTER'])

        assert api.title == 'Microsoft Planetary Computer STAC API'

    def test_invalid_url(self):
        with pytest.raises(TypeError):
            Client.open()


class TestAPISearch:
    @pytest.fixture(scope='function')
    def api(self):
        return Client.from_file(str(TEST_DATA / 'planetary-computer-root.json'))

    def test_search_conformance_error(self, api):
        """Should raise a NotImplementedError if the API doesn't conform to the Item Search spec. Message should
        include information about the spec that was not conformed to."""
        # Set the conformance to only STAC API - Core
        api._stac_io._conformance = [api._stac_io._conformance[0]]

        with pytest.raises(NotImplementedError) as excinfo:
            api.search(limit=10, max_items=10, collections='mr-peebles')
        assert str(ConformanceClasses.ITEM_SEARCH) in str(excinfo.value)

    def test_no_search_link(self, api):
        # Remove the search link
        api.remove_links('search')

        with pytest.raises(NotImplementedError) as excinfo:
            api.search(limit=10, max_items=10, collections='naip')
        assert 'No link with "rel" type of "search"' in str(excinfo.value)

    def test_search(self, api):
        results = api.search(bbox=[-73.21, 43.99, -73.12, 44.05],
                             collections='naip',
                             limit=10,
                             max_items=20,
                             datetime=[datetime(2020, 1, 1, 0, 0, 0, tzinfo=tzutc()), None])

        assert results._parameters == {
            'bbox': (-73.21, 43.99, -73.12, 44.05),
            'collections': ('naip', ),
            'limit': 10,
            'datetime': '2020-01-01T00:00:00Z/..'
        }



import argparse
import json
import logging
import os
import sys

from .client import Client
from .version import __version__

logger = logging.getLogger(__name__)


def search(client, method='GET', matched=False, save=None, **kwargs):
    """ Main function for performing a search """

    try:
        search = client.search(method=method, **kwargs)

        if matched:
            matched = search.matched()
            print('%s items matched' % matched)
        else:
            feature_collection = search.get_all_items_as_dict()
            if save:
                with open(save, 'w') as f:
                    f.write(json.dumps(feature_collection))
            else:
                print(json.dumps(feature_collection))

    except Exception as e:
        print(e)
        return 1


def collections(client, save=None, **kwargs):
    """ Fetch collections from collections endpoint """
    try:
        collections = client.get_all_collections(**kwargs)
        collections_dicts = [c.to_dict() for c in collections]
        if save:
            with open(save, 'w') as f:
                f.write(json.dumps(collections_dicts))
        else:
            print(json.dumps(collections_dicts))
    except Exception as e:
        print(e)
        return 1


def parse_args(args):
    desc = 'STAC Client'
    dhf = argparse.ArgumentDefaultsHelpFormatter
    parser0 = argparse.ArgumentParser(description=desc)
    parser0.add_argument('--version',
                         help='Print version and exit',
                         action='version',
                         version=__version__)

    parent = argparse.ArgumentParser(add_help=False)
    parent.add_argument('url', help='Root Catalog URL')
    parent.add_argument('--logging', default='INFO', help='DEBUG, INFO, WARN, ERROR, CRITICAL')
    parent.add_argument('--headers',
                        nargs='*',
                        help='Additional request headers (KEY=VALUE pairs)',
                        default=None)
    parent.add_argument('--ignore-conformance',
                        dest='ignore_conformance',
                        default=False,
                        action='store_true')

    subparsers = parser0.add_subparsers(dest='command')

    # collections command
    parser = subparsers.add_parser('collections',
                                   help='Get all collections in this Catalog',
                                   parents=[parent],
                                   formatter_class=dhf)
    output_group = parser.add_argument_group('output options')
    output_group.add_argument('--save', help='Filename to save collections to', default=None)

    # search command
    parser = subparsers.add_parser('search',
                                   help='Perform new search of items',
                                   parents=[parent],
                                   formatter_class=dhf)

    search_group = parser.add_argument_group('search options')
    search_group.add_argument('-c', '--collections', help='One or more collection IDs', nargs='*')
    search_group.add_argument('--ids',
                              help='One or more Item IDs (ignores other parameters)',
                              nargs='*')
    search_group.add_argument('--bbox',
                              help='Bounding box (min lon, min lat, max lon, max lat)',
                              nargs=4)
    search_group.add_argument('--intersects', help='GeoJSON Feature or geometry (file or string)')
    search_group.add_argument('--datetime',
                              help='Single date/time or begin and end date/time '
                              '(e.g., 2017-01-01/2017-02-15)')
    search_group.add_argument('-q',
                              '--query',
                              nargs='*',
                              help='Query properties of form '
                              'KEY=VALUE (<, >, <=, >=, = supported)')
    search_group.add_argument('--filter', help='Filter on queryables using CQL JSON')
    search_group.add_argument('--sortby', help='Sort by fields', nargs='*')
    search_group.add_argument('--fields', help='Control what fields get returned', nargs='*')
    search_group.add_argument('--limit', help='Page size limit', type=int, default=100)
    search_group.add_argument('--max-items',
                              dest='max_items',
                              help='Max items to retrieve from search',
                              type=int)
    search_group.add_argument('--method', help='GET or POST', default='POST')

    output_group = parser.add_argument_group('output options')
    output_group.add_argument('--matched',
                              help='Print number matched',
                              default=False,
                              action='store_true')
    output_group.add_argument('--save', help='Filename to save Item collection to', default=None)

    parsed_args = {k: v for k, v in vars(parser0.parse_args(args)).items() if v is not None}
    if "command" not in parsed_args:
        parser0.print_usage()
        return []

    # if intersects is JSON file, read it in
    if 'intersects' in parsed_args:
        if os.path.exists(parsed_args['intersects']):
            with open(parsed_args['intersects']) as f:
                data = json.loads(f.read())
                if data['type'] == 'Feature':
                    parsed_args['intersects'] = data['geometry']
                elif data['type'] == 'FeatureCollection':
                    parsed_args['intersects'] = data['features'][0]['geometry']
                else:
                    parsed_args['intersects'] = data

    # if headers provided, parse it
    if 'headers' in parsed_args:
        new_headers = {}
        for head in parsed_args['headers']:
            parts = head.split('=')
            if len(parts) == 2:
                new_headers[parts[0]] = parts[1]
            else:
                logger.warning(f"Unable to parse header {head}")
        parsed_args['headers'] = new_headers

    if 'filter' in parsed_args:
        parsed_args['filter'] = json.loads(parsed_args['filter'])

    return parsed_args


def cli():
    args = parse_args(sys.argv[1:])
    if not args:
        return None

    loglevel = args.pop('logging')
    if args.get('save', False) or args.get('matched', False):
        logging.basicConfig(level=loglevel)
        # quiet loggers
        logging.getLogger("urllib3").propagate = False

    cmd = args.pop('command')

    try:
        url = args.pop('url')
        headers = args.pop('headers', {})
        ignore_conformance = args.pop('ignore_conformance')
        client = Client.open(url, headers=headers, ignore_conformance=ignore_conformance)
    except Exception as e:
        print(e)
        return 1

    if cmd == 'search':
        return search(client, **args)
    elif cmd == 'collections':
        return collections(client, **args)


if __name__ == "__main__":
    return_code = cli()
    if return_code and return_code != 0:
        sys.exit(return_code)



import pytest
from pytest_console_scripts import ScriptRunner

from tests.helpers import STAC_URLS


class TestCLI:
    @pytest.mark.vcr
    def test_item_search(self, script_runner: ScriptRunner):
        args = [
            "stac-client", "search", STAC_URLS['PLANETARY-COMPUTER'], "-c", "naip", "--max-items",
            "20"
        ]
        result = script_runner.run(*args, print_result=False)
        assert result.success

    def test_no_arguments(self, script_runner: ScriptRunner):
        args = ["stac-client"]
        result = script_runner.run(*args, print_result=False)
        assert result.success



import os
from imp import load_source
from setuptools import setup, find_packages
from glob import glob

__version__ = load_source('pystac_client.version', 'pystac_client/version.py').__version__

from os.path import (
    basename,
    splitext
)

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.md')) as readme_file:
    readme = readme_file.read()

setup(
    name='pystac-client',
    version=__version__,
    description=("Python library for working with Spatiotemporal Asset Catalog (STAC)."),
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Jon Duckworth, Matthew Hanson",
    author_email='duckontheweb@gmail.com, matt.a.hanson@gmail.com',
    url='https://github.com/stac-utils/pystac-client.git',
    packages=find_packages(),
    py_modules=[splitext(basename(path))[0] for path in glob('pystac_client/*.py')],
    include_package_data=False,
    install_requires=[
        "python-dateutil>=2.7.0",
        "requests~=2.25",
        "pystac~=1.1.0"
    ],
    extras_require={
        "validation": ["jsonschema==3.2.0"]
    },
    license="Apache Software License 2.0",
    zip_safe=False,
    keywords=[
        'pystac',
        'imagery',
        'raster',
        'catalog',
        'STAC'
    ],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering :: GIS",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    test_suite='tests',
    entry_points={
        'console_scripts': ['stac-client=pystac_client.cli:cli']
    }
)



from typing import Any, Iterable, Dict, Optional, TYPE_CHECKING

import pystac
import pystac.validation
from pystac_client.collection_client import CollectionClient

from pystac_client.conformance import ConformanceClasses
from pystac_client.exceptions import APIError
from pystac_client.item_search import ItemSearch
from pystac_client.stac_api_io import StacApiIO

if TYPE_CHECKING:
    from pystac.item import Item as Item_Type


class Client(pystac.Catalog):
    """A Client for interacting with the root of a STAC Catalog or API

    Instances of the ``Client`` class inherit from :class:`pystac.Catalog` and provide a convenient way of interacting
    with STAC Catalogs OR STAC APIs that conform to the `STAC API spec <https://github.com/radiantearth/stac-api-spec>`_.
    In addition to being a valid
    `STAC Catalog <https://github.com/radiantearth/stac-spec/blob/master/catalog-spec/catalog-spec.md>`_
    APIs that have a ``"conformsTo"`` indicate that it supports additional functionality on top of a normal STAC Catalog,
    such as searching items (e.g., /search endpoint).
    """
    def __repr__(self):
        return '<Client id={}>'.format(self.id)

    @classmethod
    def open(cls,
             url: str,
             headers: Dict[str, str] = None,
             ignore_conformance: bool = False) -> "Client":
        """Opens a STAC Catalog or API
        This function will read the root catalog of a STAC Catalog or API

        Args:
            url : The URL of a STAC Catalog. If not specified, this will use the `STAC_URL` environment variable.
            headers : A dictionary of additional headers to use in all requests made to any part of this Catalog/API.
            ignore_conformance : Ignore any advertised Conformance Classes in this Catalog/API. This means that
                functions will skip checking conformance, and may throw an unknown error if that feature is
                not supported, rather than a :class:`NotImplementedError`.

        Return:
            catalog : A :class:`Client` instance for this Catalog/API
        """
        cat = cls.from_file(url, headers=headers)
        search_link = cat.get_links('search')
        # if there is a search link, but no conformsTo advertised, ignore conformance entirely
        # NOTE: this behavior to be deprecated as implementations become conformant
        if ignore_conformance or ('conformsTo' not in cat.extra_fields.keys()
                                  and len(search_link) > 0):
            cat._stac_io.set_conformance(None)
        return cat

    @classmethod
    def from_file(cls,
                  href: str,
                  stac_io: Optional[pystac.StacIO] = None,
                  headers: Optional[Dict] = {}) -> "Client":
        """Open a STAC Catalog/API

        Returns:
            Client: A Client (PySTAC Catalog) of the root Catalog for this Catalog/API
        """
        if stac_io is None:
            stac_io = StacApiIO(headers=headers)

        cat = super().from_file(href, stac_io)

        cat._stac_io._conformance = cat.extra_fields.get('conformsTo', [])

        return cat

    def get_collection(self, collection_id: str) -> CollectionClient:
        """Get a single collection from this Catalog/API

        Args:
            collection_id: The Collection ID to get

        Returns:
            CollectionClient: A STAC Collection
        """
        for col in self.get_collections():
            if col.id == collection_id:
                return col

    def get_collections(self) -> Iterable[CollectionClient]:
        """ Get Collections in this Catalog

            Gets the collections from the /collections endpoint if supported, otherwise fall
            back to Catalog behavior of following child links

        Return:
            Iterable[CollectionClient]: Iterator through Collections in Catalog/API
        """
        if self._stac_io.assert_conforms_to(ConformanceClasses.COLLECTIONS):
            url = self.get_self_href() + '/collections'
            for page in self._stac_io.get_pages(url):
                if 'collections' not in page:
                    raise APIError("Invalid response from /collections")
                for col in page['collections']:
                    collection = CollectionClient.from_dict(col, root=self)
                    yield collection
        else:
            yield from super().get_collections()

    def get_items(self) -> Iterable["Item_Type"]:
        """Return all items of this catalog.

        Return:
            Iterable[Item]:: Generator of items whose parent is this catalog.
        """
        if self._stac_io.assert_conforms_to(ConformanceClasses.ITEM_SEARCH):
            search = self.search()
            yield from search.get_items()
        else:
            return super().get_items()

    def get_all_items(self) -> Iterable["Item_Type"]:
        """Get all items from this catalog and all subcatalogs. Will traverse
        any subcatalogs recursively, or use the /search endpoint if supported

        Returns:
            Iterable[Item]:: All items that belong to this catalog, and all
                catalogs or collections connected to this catalog through
                child links.
        """
        if self._stac_io.assert_conforms_to(ConformanceClasses.ITEM_SEARCH):
            yield from self.get_items()
        else:
            yield from super().get_items()

    def search(self, **kwargs: Any) -> ItemSearch:
        """Query the ``/search`` endpoint using the given parameters.

        This method returns an :class:`~pystac_client.ItemSearch` instance, see that class's documentation
        for details on how to get the number of matches and iterate over results. All keyword arguments are passed
        directly to the :class:`~pystac_client.ItemSearch` instance.

        .. warning::

            This method is only implemented if the API conforms to the
            `STAC API - Item Search <https://github.com/radiantearth/stac-api-spec/tree/master/item-search>`__ spec
            *and* contains a link with a ``"rel"`` type of ``"search"`` in its root catalog.
            If the API does not meet either of these criteria, this method will raise a :exc:`NotImplementedError`.

        Args:
            **kwargs : Any parameter to the :class:`~pystac_client.ItemSearch` class, other than `url`, `conformance`,
                and `stac_io` which are set from this Client instance

        Returns:
            search : An ItemSearch instance that can be used to iterate through Items.

        Raises:
            NotImplementedError: If the API does not conform to the `Item Search spec
                <https://github.com/radiantearth/stac-api-spec/tree/master/item-search>`__ or does not have a link with
                a ``"rel"`` type of ``"search"``.
        """

        search_link = self.get_single_link('search')
        if search_link is None:
            raise NotImplementedError(
                'No link with "rel" type of "search" could be found in this catalog')

        return ItemSearch(search_link.target, stac_io=self._stac_io, client=self, **kwargs)



from dateutil.tz import tzutc
from dateutil.relativedelta import relativedelta
import json
import re
from collections.abc import Iterable, Mapping
from copy import deepcopy
from datetime import timezone, datetime as datetime_
from typing import Dict, Iterator, List, Optional, TYPE_CHECKING, Tuple, Union
import warnings

from pystac import Collection, Item, ItemCollection
from pystac.stac_io import StacIO

from pystac_client.stac_api_io import StacApiIO
from pystac_client.conformance import ConformanceClasses

if TYPE_CHECKING:
    from pystac_client.client import Client

DATETIME_REGEX = re.compile(r"(?P<year>\d{4})(\-(?P<month>\d{2})(\-(?P<day>\d{2})"
                            r"(?P<remainder>(T|t)\d{2}:\d{2}:\d{2}(\.\d+)?"
                            r"(?P<tz_info>Z|([-+])(\d{2}):(\d{2}))?)?)?)?")

DatetimeOrTimestamp = Optional[Union[datetime_, str]]
Datetime = Union[Tuple[str], Tuple[str, str]]
DatetimeLike = Union[DatetimeOrTimestamp, Tuple[DatetimeOrTimestamp, DatetimeOrTimestamp],
                     List[DatetimeOrTimestamp], Iterator[DatetimeOrTimestamp]]

BBox = Tuple[float, ...]
BBoxLike = Union[BBox, List[float], Iterator[float], str]

Collections = Tuple[str, ...]
CollectionsLike = Union[List[str], Iterator[str], str]

IDs = Tuple[str, ...]
IDsLike = Union[IDs, str, List[str], Iterator[str]]

Intersects = dict
IntersectsLike = Union[str, Intersects, object]

Query = dict
QueryLike = Union[Query, List[str]]

FilterLike = dict

Sortby = List[str]
SortbyLike = Union[Sortby, str]

Fields = List[str]
FieldsLike = Union[Fields, str]


# from https://gist.github.com/angstwad/bf22d1822c38a92ec0a9#gistcomment-2622319
def dict_merge(dct: Dict, merge_dct: Dict, add_keys: bool = True) -> Dict:
    """ Recursive dict merge.

    Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``. This version will return a copy of the dictionary and leave the original
    arguments untouched.  The optional argument ``add_keys``, determines whether keys which are
    present in ``merge_dict`` but not ``dct`` should be included in the new dict.

    Args:
        dct (dict) onto which the merge is executed
        merge_dct (dict): dct merged into dct
        add_keys (bool): whether to add new keys

    Return:
        dict: updated dict
    """
    dct = dct.copy()
    if not add_keys:
        merge_dct = {k: merge_dct[k] for k in set(dct).intersection(set(merge_dct))}

    for k, v in merge_dct.items():
        if (k in dct and isinstance(dct[k], dict) and isinstance(merge_dct[k], Mapping)):
            dct[k] = dict_merge(dct[k], merge_dct[k], add_keys=add_keys)
        else:
            dct[k] = merge_dct[k]

    return dct


class ItemSearch:
    """Represents a deferred query to a STAC search endpoint as described in the
    `STAC API - Item Search spec <https://github.com/radiantearth/stac-api-spec/tree/master/item-search>`__.

    No request is sent to the API until a function is called to fetch or iterate through the resulting STAC Items,
     either the :meth:`ItemSearch.item_collections` or :meth:`ItemSearch.items` method is called and iterated over.

    All "Parameters", with the exception of ``max_items``, ``method``, and ``url`` correspond to query parameters
    described in the `STAC API - Item Search: Query Parameters Table
    <https://github.com/radiantearth/stac-api-spec/tree/master/item-search#query-parameter-table>`__ docs. Please refer
    to those docs for details on how these parameters filter search results.

    Args:
        url : The URL to the item-search endpoint
        method : The HTTP method to use when making a request to the service. This must be either ``"GET"``, ``"POST"``, or
            ``None``. If ``None``, this will default to ``"POST"`` if the ``intersects`` argument is present and ``"GET"``
            if not. If a ``"POST"`` request receives a ``405`` status for the response, it will automatically retry with a
            ``"GET"`` request for all subsequent requests.
        max_items : The maximum number of items to return from the search. *Note that this is not a STAC API - Item Search
            parameter and is instead used by the client to limit the total number of returned items*.
        limit : The maximum number of items to return *per page*. Defaults to ``None``, which falls back to the limit set
            by the service.
        bbox: May be a list, tuple, or iterator representing a bounding box of 2D or 3D coordinates. Results will be filtered
            to only those intersecting the bounding box.
        datetime: Either a single datetime or datetime range used to filter results. You may express a single datetime
            using a :class:`datetime.datetime` instance, a `RFC 3339-compliant <https://tools.ietf.org/html/rfc3339>`__
            timestamp, or a simple date string (see below). Instances of :class:`datetime.datetime` may be either
            timezone aware or unaware. Timezone aware instances will be converted to a UTC timestamp before being passed
            to the endpoint. Timezone unaware instances are assumed to represent UTC timestamps. You may represent a
            datetime range using a ``"/"`` separated string as described in the spec, or a list, tuple, or iterator
            of 2 timestamps or datetime instances. For open-ended ranges, use either ``".."`` (``'2020-01-01:00:00:00Z/..'``,
            ``['2020-01-01:00:00:00Z', '..']``) or a value of ``None`` (``['2020-01-01:00:00:00Z', None]``).

            If using a simple date string, the datetime can be specified in ``YYYY-mm-dd`` format, optionally truncating
            to ``YYYY-mm`` or just ``YYYY``. Simple date strings will be expanded to include the entire time period, for
            example:

            - ``2017`` expands to ``2017-01-01T00:00:00Z/2017-12-31T23:59:59Z``
            - ``2017-06`` expands to ``2017-06-01T00:00:00Z/2017-06-30T23:59:59Z``
            - ``2017-06-10`` expands to ``2017-06-10T00:00:00Z/2017-06-10T23:59:59Z``

            If used in a range, the end of the range expands to the end of that day/month/year, for example:

            - ``2017/2018`` expands to ``2017-01-01T00:00:00Z/2018-12-31T23:59:59Z``
            - ``2017-06/2017-07`` expands to ``2017-06-01T00:00:00Z/2017-07-31T23:59:59Z``
            - ``2017-06-10/2017-06-11`` expands to ``2017-06-10T00:00:00Z/2017-06-11T23:59:59Z``
        intersects: A GeoJSON-like dictionary or JSON string. Results filtered to only those intersecting the geometry
        ids: List of Item ids to return. All other filter parameters that further restrict the number of search results
            (except ``limit``) are ignored.
        collections: List of one or more Collection IDs or :class:`pystac.Collection` instances. Only Items in one
            of the provided Collections will be searched
        query: List or JSON of query parameters as per the STAC API `query` extension
        filter: JSON of query parameters as per the STAC API `filter` extension
        sortby: A single field or list of fields to sort the response by
        fields: A list of fields to return in the response. Note this may result in invalid JSON.
            Use `get_all_items_as_dict` to avoid errors
        max_items: The maximum number of items to get, even if there are more matched items
        method: The http method, 'GET' or 'POST'
        stac_io: An instance of of StacIO for retrieving results. Normally comes from the Client that returns this ItemSearch
        client: An instance of a root Client used to set the root on resulting Items
    """
    def __init__(self,
                 url: str,
                 *,
                 limit: Optional[int] = None,
                 bbox: Optional[BBoxLike] = None,
                 datetime: Optional[DatetimeLike] = None,
                 intersects: Optional[IntersectsLike] = None,
                 ids: Optional[IDsLike] = None,
                 collections: Optional[CollectionsLike] = None,
                 query: Optional[QueryLike] = None,
                 filter: Optional[FilterLike] = None,
                 sortby: Optional[SortbyLike] = None,
                 fields: Optional[FieldsLike] = None,
                 max_items: Optional[int] = None,
                 method: Optional[str] = 'POST',
                 stac_io: Optional[StacIO] = None,
                 client: Optional["Client"] = None):
        self.url = url
        self.client = client

        if stac_io:
            self._stac_io = stac_io
        else:
            self._stac_io = StacApiIO()
        self._stac_io.assert_conforms_to(ConformanceClasses.ITEM_SEARCH)

        self._max_items = max_items
        if self._max_items is not None and limit is not None:
            limit = min(limit, self._max_items)

        if limit is not None and (limit < 1 or limit > 10000):
            raise Exception(f"Invalid limit of {limit}, must be between 1 and 10,000")

        self.method = method

        params = {
            'limit': limit,
            'bbox': self._format_bbox(bbox),
            'datetime': self._format_datetime(datetime),
            'ids': self._format_ids(ids),
            'collections': self._format_collections(collections),
            'intersects': self._format_intersects(intersects),
            'query': self._format_query(query),
            'filter': self._format_filter(filter),
            'sortby': self._format_sortby(sortby),
            'fields': self._format_fields(fields)
        }

        if params['filter'] is not None:
            params['filter-lang'] = 'cql-json'

        self._parameters = {k: v for k, v in params.items() if v is not None}

    def get_parameters(self):
        if self.method == 'POST':
            return self._parameters
        elif self.method == 'GET':
            params = deepcopy(self._parameters)
            if 'bbox' in params:
                params['bbox'] = ','.join(params['bbox'])
            if 'ids' in params:
                params['ids'] = ','.join(params['ids'])
            if 'collections' in params:
                params['collections'] = ','.join(params['collections'])
            if 'intersects' in params:
                params['intersects'] = json.dumps(params['intersects'])
            return params
        else:
            raise Exception(f"Unsupported method {self.method}")

    @staticmethod
    def _format_query(value: List[QueryLike]) -> Optional[dict]:
        if value is None:
            return None

        OP_MAP = {'>=': 'gte', '<=': 'lte', '=': 'eq', '>': 'gt', '<': 'lt'}

        if isinstance(value, list):
            query = {}
            for q in value:
                for op in ['>=', '<=', '=', '>', '<']:
                    parts = q.split(op)
                    if len(parts) == 2:
                        param = parts[0]
                        val = parts[1]
                        if param == "gsd":
                            val = float(val)
                        query = dict_merge(query, {parts[0]: {OP_MAP[op]: val}})
                        break
        else:
            query = value

        return query

    def _format_filter(self, value: FilterLike) -> Optional[dict]:
        if value is None:
            return None

        self._stac_io.assert_conforms_to(ConformanceClasses.FILTER)
        return value

    @staticmethod
    def _format_bbox(value: Optional[BBoxLike]) -> Optional[BBox]:
        if value is None:
            return None

        if isinstance(value, str):
            bbox = tuple(map(float, value.split(',')))
        else:
            bbox = tuple(map(float, value))

        return bbox

    @staticmethod
    def _format_datetime(value: Optional[DatetimeLike]) -> Optional[Datetime]:
        def _to_utc_isoformat(dt):
            dt = dt.astimezone(timezone.utc)
            dt = dt.replace(tzinfo=None)
            return dt.isoformat("T") + "Z"

        def _to_isoformat_range(component: DatetimeOrTimestamp):
            """Converts a single DatetimeOrTimestamp into one or two Datetimes.

            This is required to expand a single value like "2017" out to the whole year. This function returns two values.
            The first value is always a valid Datetime. The second value can be None or a Datetime. If it is None, this
            means that the first value was an exactly specified value (e.g. a `datetime.datetime`). If the second value is
            a Datetime, then it will be the end of the range at the resolution of the component, e.g. if the component
            were "2017" the second value would be the last second of the last day of 2017.
            """
            if component is None:
                return "..", None
            elif isinstance(component, str):
                if component == "..":
                    return component, None

                match = DATETIME_REGEX.match(component)
                if not match:
                    raise Exception(f"invalid datetime component: {component}")
                elif match.group("remainder"):
                    if match.group("tz_info"):
                        return component, None
                    else:
                        return f"{component}Z", None
                else:
                    year = int(match.group("year"))
                    optional_month = match.group("month")
                    optional_day = match.group("day")

                if optional_day is not None:
                    start = datetime_(year,
                                      int(optional_month),
                                      int(optional_day),
                                      0,
                                      0,
                                      0,
                                      tzinfo=tzutc())
                    end = start + relativedelta(days=1, seconds=-1)
                elif optional_month is not None:
                    start = datetime_(year, int(optional_month), 1, 0, 0, 0, tzinfo=tzutc())
                    end = start + relativedelta(months=1, seconds=-1)
                else:
                    start = datetime_(year, 1, 1, 0, 0, 0, tzinfo=tzutc())
                    end = start + relativedelta(years=1, seconds=-1)
                return _to_utc_isoformat(start), _to_utc_isoformat(end)
            else:
                return _to_utc_isoformat(component), None

        if value is None:
            return None
        elif isinstance(value, datetime_):
            return _to_utc_isoformat(value)
        elif isinstance(value, str):
            components = value.split("/")
        else:
            components = list(value)

        if not components:
            return None
        elif len(components) == 1:
            start, end = _to_isoformat_range(components[0])
            if end is not None:
                return f"{start}/{end}"
            else:
                return start
        elif len(components) == 2:
            start, _ = _to_isoformat_range(components[0])
            backup_end, end = _to_isoformat_range(components[1])
            return f"{start}/{end or backup_end}"
        else:
            raise Exception(
                f"too many datetime components (max=2, actual={len(components)}): {value}")

    @staticmethod
    def _format_collections(value: Optional[CollectionsLike]) -> Optional[Collections]:
        def _format(c):
            if isinstance(c, str):
                return c
            if isinstance(c, Iterable):
                return tuple(map(_format, c))

            return c.id

        if value is None:
            return None
        if isinstance(value, str):
            return tuple(map(_format, value.split(',')))
        if isinstance(value, Collection):
            return _format(value),

        return _format(value)

    @staticmethod
    def _format_ids(value: Optional[IDsLike]) -> Optional[IDs]:
        if value is None:
            return None

        if isinstance(value, str):
            return tuple(value.split(','))

        return tuple(value)

    def _format_sortby(self, value: Optional[SortbyLike]) -> Optional[Sortby]:
        if value is None:
            return None

        self._stac_io.assert_conforms_to(ConformanceClasses.SORT)

        if isinstance(value, str):
            return tuple(value.split(','))

        return tuple(value)

    def _format_fields(self, value: Optional[FieldsLike]) -> Optional[Fields]:
        if value is None:
            return None

        self._stac_io.assert_conforms_to(ConformanceClasses.FIELDS)

        if isinstance(value, str):
            return tuple(value.split(','))

        return tuple(value)

    @staticmethod
    def _format_intersects(value: Optional[IntersectsLike]) -> Optional[Intersects]:
        if value is None:
            return None
        if isinstance(value, str):
            return json.loads(value)
        return deepcopy(getattr(value, '__geo_interface__', value))

    def matched(self) -> int:
        """Return number matched for search

        Returns the value from the `numberMatched` or `context.matched` field. Not all APIs
        will support counts in which case a warning will be issued

        Returns:
            int: Total count of matched items. If counts are not supported `None` is returned.
        """
        params = {**self.get_parameters(), "limit": 1}
        resp = self._stac_io.read_json(self.url, method=self.method, parameters=params)
        found = None
        if 'context' in resp:
            found = resp['context']['matched']
        elif 'numberMatched' in resp:
            found = resp['numberMatched']
        if found is None:
            warnings.warn("numberMatched or context.matched not in response")
        return found

    def get_item_collections(self) -> Iterator[ItemCollection]:
        """Iterator that yields ItemCollection objects.  Each ItemCollection is a page of results
        from the search.

        Yields:
            Iterable[Item] : pystac_client.ItemCollection
        """
        for page in self._stac_io.get_pages(self.url, self.method, self.get_parameters()):
            yield ItemCollection.from_dict(page, preserve_dict=False, root=self.client)

    def get_items(self) -> Iterator[Item]:
        """Iterator that yields :class:`pystac.Item` instances for each item matching the given search parameters. Calls
        :meth:`ItemSearch.item_collections()` internally and yields from
        :attr:`ItemCollection.features <pystac_client.ItemCollection.features>` for each page of results.

        Return:
            Iterable[Item] : Iterate through resulting Items
        """
        nitems = 0
        for item_collection in self.get_item_collections():
            for item in item_collection:
                yield item
                nitems += 1
                if self._max_items and nitems >= self._max_items:
                    return

    def get_all_items_as_dict(self) -> Dict:
        """Convenience method that gets all items from all pages, up to self._max_items,
         and returns an array of dictionaries

        Return:
            Dict : A GeoJSON FeatureCollection
        """
        features = []
        for page in self._stac_io.get_pages(self.url, self.method, self.get_parameters()):
            for feature in page['features']:
                features.append(feature)
                if self._max_items and len(features) >= self._max_items:
                    return {"type": "FeatureCollection", "features": features}
        return {"type": "FeatureCollection", "features": features}

    def get_all_items(self) -> ItemCollection:
        """Convenience method that builds an :class:`ItemCollection` from all items matching the given search parameters.

        Return:
            item_collection : ItemCollection
        """
        feature_collection = self.get_all_items_as_dict()
        return ItemCollection.from_dict(feature_collection, preserve_dict=False, root=self.client)



import json
from pathlib import Path

TEST_DATA = Path(__file__).parent / 'data'

STAC_URLS = {
    "PLANETARY-COMPUTER": "https://planetarycomputer.microsoft.com/api/stac/v1",
    "EARTH-SEARCH": "https://earth-search.aws.element84.com/v0",
    "MLHUB": "https://api.radiant.earth/mlhub/v1",
}


def read_data_file(file_name: str, mode='r', parse_json=False):
    file_path = TEST_DATA / file_name
    with file_path.open(mode=mode) as src:
        if parse_json:
            return json.load(src)
        else:
            return src.read()
